import re
from pathlib import Path

import httpx
import pytest
import yaml
from referencing import Registry, Resource
from referencing.jsonschema import DRAFT202012

from .utils import (
    get_collection_responses,
    get_links_from_header,
    get_features_responses,
    get_feature_responses,
)


def pytest_collection_modifyitems(items):
    # Run the collected test functions in order of Abstract test number
    items[:] = sorted(items, key=lambda item: int(item.name.replace("test_ast", "")))


@pytest.fixture(scope="session")
def paging_limit():
    return 3


@pytest.fixture(scope="session")
def collections_limit():
    return 20


@pytest.fixture(scope="session")
def schema_dir():
    return Path(__file__).parent.parent / "schemas"


@pytest.fixture(scope="session")
def landing_page(http_client: httpx.Client) -> httpx.Response:
    return http_client.get("/")


@pytest.fixture(scope="session")
def api_links(landing_page: httpx.Response) -> list[dict]:
    try:
        api_links = [
            link
            for link in landing_page.json()["links"]
            if link["rel"] == "service-desc" or link["rel"] == "service-doc"
        ]
    except KeyError:
        try:
            links = get_links_from_header(landing_page.headers)
            api_links = [
                link
                for link in links
                if link["rel"] == ["service-doc"] or link["rel"] == ["service-desc"]
            ]
        except KeyError:
            return []

    # HACK: because the demo.ldproxy.net test instances have a bug
    for i, link in enumerate(api_links):
        link["href"] = re.sub(r"(?<!:)//", "/", link["href"])
        api_links[i] = link

    return api_links


@pytest.fixture(scope="session")
def api_model(http_client: httpx.Client, api_links: list) -> dict:
    api_model = {}
    for link in api_links:
        # open api definition in json
        if link["type"] == "application/vnd.oai.openapi+json;version=3.0":
            response = httpx.get(link["href"])
            api_model = response.json()
        # open api definition in yaml
        elif link["type"] == "application/vnd.oai.openapi;version=3.0":
            response = httpx.get(link["href"])
            api_model = yaml.safe_load(response.content)
    return api_model


@pytest.fixture(scope="session")
def conformance_page(http_client: httpx.Client) -> httpx.Response:
    return http_client.get("/conformance")


@pytest.fixture(scope="session")
def conformance_classes(http_client: httpx.Client) -> list:
    response = http_client.get("/conformance")
    return response.json().get("conformsTo", [])


@pytest.fixture(scope="session")
def collections_page(http_client: httpx.Client) -> httpx.Response:
    return http_client.get("/collections", headers={"Accept": "application/json"})


@pytest.fixture(scope="session")
def collections(collections_page: httpx.Response, collections_limit: int) -> list[dict]:
    collections = collections_page.json().get("collections", [])
    if len(collections) > collections_limit:
        collections = collections[:collections_limit]
    return collections


@pytest.fixture(scope="session")
def collection_responses(
    collections: list[dict], http_client: httpx.Client
) -> list[dict]:
    return get_collection_responses(collections, http_client)


@pytest.fixture(scope="session")
def features_responses(
    collections: list[dict], http_client: httpx.Client
) -> list[dict]:
    return get_features_responses(collections, http_client)


@pytest.fixture(scope="module")
def feature_responses(features_responses: list[dict], http_client) -> list[dict]:
    features_limit = 100
    return get_feature_responses(features_responses, http_client, features_limit)


@pytest.fixture(scope="session")
def registry(schema_dir: Path):
    files = schema_dir.glob("*.yaml")
    resources = []
    for file in files:
        with open(file) as f:
            schema = yaml.safe_load(f)
        resource = Resource.from_contents(schema, default_specification=DRAFT202012)
        resources.append((f"urn:{file.name.replace('.yaml', '')}", resource))
    registry = Registry().with_resources(resources)
    return registry



import json

import geopandas as gpd


def test_ast2(conformance_classes: list):
    """/ats/core/crs84
    Test Purpose
        Validate that all spatial geometries provided through the API are in the CRS84 spatial reference system unless
        otherwise requested by the client.

    Test Method
        - Do not specify a coordinate reference system in any request.
            All spatial data should be in the CRS84 reference system.
        - Validate retrieved spatial data using the CRS84 reference system.
    """
    crs_conformance_class = "http://www.opengis.net/spec/ogcapi-features-2/1.0/conf/crs"
    assert (
        crs_conformance_class in conformance_classes
    ), f"OGC Features API CRS conformance class {crs_conformance_class} is not listed in the conformsTo property."


def assert_valid_feature_crs(feature: dict):
    parsed_feature = gpd.read_file(json.dumps(feature), driver="GeoJSON")
    assert parsed_feature.crs.name.startswith(
        "WGS 84"
    ), f"feature not in CRS format. expected WGS84 or WGS84H got {parsed_feature.crs.name}"



import os
from datetime import datetime
import argparse
import inspect
from importlib import import_module
from pathlib import Path

import pytest


root_dir = Path(__file__).parent
test_files = root_dir.glob("features/core/test_*.py")
epilog = """
features/
  core/
"""
for file in test_files:
    epilog += "    " + str(file.name) + "::\n"
    module = import_module(
        "." + str(file)[str(file).find("features"):].replace("/", ".").replace(".py", ""),
        package="ogctests",
    )
    funcs = [
        func[0]
        for func in inspect.getmembers(module, inspect.isfunction)
        if func[0].startswith("test")
    ]
    for func in funcs:
        epilog += "      " + func + "\n"

parser = argparse.ArgumentParser(
    prog="ogctests",
    description="OGC API Test Suite",
    formatter_class=argparse.RawDescriptionHelpFormatter,
    epilog=epilog,
)
parser.add_argument(
    "scope", type=str, help="Specify tests to run", nargs="+", default="features/core"
)
parser.add_argument(
    "-i",
    "--instance-url",
    type=str,
    help="URL of the server to run the test suite against.",
    required=True,
    dest="iurl",
)
parser.add_argument(
    "-r",
    "--report",
    type=bool,
    help="set True to enable report generation.",
    default=False,
    dest="report",
)
parser.add_argument(
    "-v",
    "--verbose",
    type=bool,
    help="set True to for verbose output",
    dest="verbose"
)
args = parser.parse_args()
os.environ["INSTANCE_URL"] = args.iurl
arglist = [str(root_dir / path) for path in args.scope]
if not args.verbose:
    arglist += ["--no-header", "--no-summary", "-p no:warnings", "--tb=no"]
if args.report:
    report_folder = os.path.expanduser("~") + "/ogctestsReporting"
    report_name = f"testrun-{datetime.now():%Y-%m-%d}.xml"
    if not os.path.exists(report_folder):
        os.mkdir(report_folder)
    arglist += [f"--junit-xml={report_folder}/{report_name}"]
print(f"\n\nTesting endpoint: {args.iurl}")
pytest.main(args=arglist)
if args.report:
    print(f"Report saved to {report_folder}/{report_name}\n\n")



import httpx


def test_ast20(collections: list[dict], http_client: httpx.Client):
    """/ats/core/query-param-invalid
    Test Purpose
        Validate that the API correctly deals with invalid query parameters.

    Test Method
        - Enter an HTTP request with an invalid query parameter.
        - Verify that the API returns the status code 400.
    """
    for collection in collections:
        path = f"/collections/{collection['id']}/items"
        param = {"limit": "unlimited"}
        response = http_client.get(path, params=param)
        assert (
            response.status_code == 400
        ), f"{path} did not respond 400 to an invalid query parameter: {param}"


def test_ast21(collections: list[dict], http_client: httpx.Client):
    """/ats/core/query-param-unknown
    Test Purpose
        Validate that the API correctly deals with unknown query parameters.

    Test Method
        - Enter an HTTP request with an query parameter that is not specified in the API definition.
        - Verify that the API returns the status code 400.
    """
    for collection in collections:
        path = f"/collections/{collection['id']}/items"
        param = {"ooga": "booga"}
        response = http_client.get(path, params=param)
        assert (
            response.status_code == 400
        ), f"{path} did not respond 400 to an unknown query parameter: {param}"



import warnings
from pathlib import Path

import httpx
import pytest
import yaml
from jsonschema.exceptions import ValidationError

from referencing import Resource, Registry
from openapi_schema_validator import validate

from .utils import get_links_by_rel


def test_ast3(http_client: httpx.Client):
    """/ats/core/root-op
    Test Purpose
        Validate that a landing page can be retrieved from the expected location.

    Test Method
        - Issue an HTTP GET request to the URL {root}/
        - Validate that a document was returned with a status code 200
        - Validate the contents of the returned document using test /ats/core/root-success.
    """
    response = http_client.get("/", headers={"Accept": "application/json"})
    assert response.status_code != 405, "Get operation not supported at /"
    assert (
        response.status_code == 200
    ), "Response was not returned with status code 200"


def test_ast4(landing_page: httpx.Response, schema_dir: Path, registry: Registry, http_client):
    """/ats/core/root-success
    Test Purpose
        Validate that the landing page complies with the require structure and contents.

    Test Method
        - Validate the landing page for all supported media types using the resources and tests identified in
          Schema and Tests for Landing Pages
        - For formats that require manual inspection, perform the following:
        - Validate that the landing page includes a "service-desc" and/or "service-doc" link to an API Definition
        - Validate that the landing page includes a "conformance" link to the conformance class declaration
        - Validate that the landing page includes a "data" link to the Feature contents.
    """
    if landing_page.headers.get("Content-Type") != "application/json":
        landing_page = http_client.get("/", headers={"Accept": "application/json"})
        if landing_page.status_code != 200 or landing_page.headers.get("Content-Type") != "application/json":
            pytest.skip("The landing page could not be retrieved in JSON. Manual inspection required.")

    with open(schema_dir / "landingPage.yaml") as file:
        schema = yaml.safe_load(file)

    validate(landing_page.json(), schema, registry=registry)

    service_desc = get_links_by_rel(landing_page, "service-desc")
    service_doc = get_links_by_rel(landing_page, "service-doc")
    conformance = get_links_by_rel(landing_page, "conformance")
    data = get_links_by_rel(landing_page, "data")

    assert any([service_desc, service_doc]) and all(
        [conformance, data]
    ), """The landing page must include at least links with relation type ('service-desc' or 'service-doc'),\
        and 'conformance' and 'data'\
        """



from pathlib import Path

import httpx
import yaml
from openapi_schema_validator import validate, OAS30Validator


def test_ast5(http_client: httpx.Client, api_links: list[dict]):
    """/ats/core/api-definition-op
    Test Purpose
        Validate that the API Definition document can be retrieved from the expected location.
    Test Method
        - Construct a path for each API Definition link on the landing page
        - Issue a HTTP GET request on each path
        - Validate that a document was returned with a status code 200
        - Validate the contents of the returned document using test /ats/core/api-definition-success.
    """
    assert api_links != [], "No API Definitions provided by the server"
    for api_link in api_links:
        response = http_client.get(api_link["href"])
        assert (
            response.status_code == 200
        ), f"Listed API Definition unavailable ({api_link['href']})"


def test_ast6(schema_dir: Path, api_model: dict):
    """/ats/core/api-definition-success
    Test Purpose
        Validate that the API Definition complies with the required structure and contents.
    Test Method
        Validate the API Definition document against an appropriate schema document.
    """
    with open(schema_dir / "openapi_schema_v3.0.yaml", "r") as file:
        schema = yaml.safe_load(file)
    validate(api_model, schema, cls=OAS30Validator)



import httpx


def parse_link_header(raw_link_header: str) -> list[dict]:
    """Convert a link header to a list of dictionaries"""
    if not raw_link_header:
        return []
    links = raw_link_header.split(",")
    parsed_links = []
    for link in links:
        parts = link.split(";")
        parsed_link = {}
        for part in parts:
            part = part.strip()
            if part.startswith("<") and part.endswith(">"):
                parsed_link["iri"] = part
            else:
                k, v = part.split("=")
                parsed_link[k.strip()] = v.strip()
        parsed_links.append(parsed_link)
    return parsed_links


def get_supported_media_types(response: httpx.Response) -> set:
    """Get the list of media types supported at /collections/{collectionId}/items

    Relies on definition of self and alternate links in the response headers
    """
    links = parse_link_header(response.headers.get("link", ""))
    return {
        link.get("type", None).strip('"') for link in links if link.get("rel").strip('"') == "self" or link.get("rel").strip('"') == "alternate"
    }



import warnings

import pytest
import httpx
import yaml

from datetime import datetime, timezone, timedelta


@pytest.fixture(scope="session")
def features_limit():
    return 100


def get_links_from_header(headers: dict) -> dict:
    links = {}
    link_strings = headers["link"]
    for link_string in link_strings:
        fragments = link_string.split(",")
        for i, fragment in enumerate(fragments):
            links[i] = {}
            fragment.strip()
            if fragment.startswith("<") and fragment.endswith(">"):
                links[i]["href"] = fragment[1:-1]
            else:
                k, v = fragment.split("=").strip('"')
                links[i][k] = v
    return links


def get_links_by_rel(
    response: httpx.Response, rel: str, only_first: bool = False
) -> list | dict:
    """get all the links where link["rel"] == rel

    optionally only return the first link to meet the criteria
    """
    links = response.json().get("links", [])
    filtered = [link for link in links if link.get("rel") == rel]
    if only_first and len(filtered) > 0:
        return filtered[0]
    return filtered


def check_links_rel_type(response: httpx.Response) -> bool:
    """Check if all links specify the rel and type parameters"""
    links = response.json().get("links", [])
    if not links:
        return False
    for link in links:
        if link.get("rel", None) is None or link.get("type", None) is None:
            return False
    return True


def resolve_json_schema_ref(model: dict, reference: str) -> dict:
    """assumes ref of type: [http.*.(yaml|json)]#/.../..."""
    url = reference[0 : reference.find("#")] if reference.startswith("http") else ""
    if url:
        response = httpx.get(url)
        if response.headers.get("Content-Type") == "application/json":
            model = response.json()
        elif "yaml" in response.headers.get("Content-Type") or response.headers.get(
            "Content-Type"
        ).startswith("text/plain"):
            try:
                model = yaml.safe_load(response.content)
            except yaml.YAMLError:
                warnings.warn(f"could not decode json schema $ref at {url}")
                return {}
    reference_fragments = reference[reference.find("#") :].split("/")
    obj = model
    for fragment in reference_fragments[1:]:
        try:
            obj = obj[fragment]
        except KeyError:
            warnings.warn(
                f"Could not resolve the json schema $ref {reference} in the api model."
            )
            return {}
    return obj


def get_api_parameter(model: dict, path: str, parameter_name: str) -> dict:
    try:
        path_params = model["paths"][path]["parameters"]
    except KeyError:
        path_params = []
    for param in path_params:
        if "$ref" in param.keys():
            param = resolve_json_schema_ref(model, param["$ref"])
        if param.get("name") == parameter_name:
            return param
    try:
        get_params = model["paths"][path]["get"]["parameters"]
    except KeyError:
        get_params = []
    for param in get_params:
        if "$ref" in param.keys():
            param = resolve_json_schema_ref(model, param["$ref"])
        if param.get("name") == parameter_name:
            return param
    return {}


def get_collection_responses(
    collections: list[dict], http_client: httpx.Client
) -> list[dict]:
    responses = []
    for collection in collections:
        collectionId = collection["id"]
        path = f"/collections/{collectionId}"
        response = dict()
        response["id"] = collectionId
        response["response"] = http_client.get(path)
        responses.append(response)
    return responses


def get_features_responses(
    collections: list[dict],
    http_client: httpx.Client,
    headers: dict = {"Accept": "application/geo+json"},
    params: dict = None,
) -> list[dict]:
    responses = []
    for collection in collections:
        collectionId = collection["id"]
        path = f"/collections/{collectionId}/items"
        response = dict()
        response["id"] = collectionId
        response["before"] = datetime.now(tz=timezone.utc) - timedelta(seconds=1)
        response["response"] = http_client.get(path, headers=headers, params=params)
        response["after"] = datetime.now(tz=timezone.utc) + timedelta(seconds=1)
        responses.append(response)
    return responses


def get_feature_responses(
    features_responses: list[dict], http_client: httpx.Client, features_limit: int
) -> list[dict]:
    feature_responses = []
    for response in features_responses:
        features = response["response"].json().get("features", [])
        if not features:
            warnings.warn(f"No features found in collection: {response['id']}")
            break
        for i, feature in enumerate(features):
            if i > features_limit:
                break
            path = f"/collections/{response['id']}/items/{feature['id']}"
            feature_response = dict(
                {"feature_id": feature["id"], "collection_id": response["id"]}
            )
            feature_response["response"] = http_client.get(
                path, headers={"Accept": "application/geo+json"}
            )
            feature_responses.append(feature_response)
    return feature_responses



from pathlib import Path

import httpx
import yaml
from openapi_schema_validator import validate


def test_ast7(conformance_page: httpx.Response):
    """/ats/core/conformance-op
    Test Purpose
        Validate that a Conformance Declaration can be retrieved from the expected location.

    Test Method
        - Construct a path for each "conformance" link on the landing page as well as for the {root}/conformance path.
        - Issue an HTTP GET request on each path
        - Validate that a document was returned with a status code 200
        - Validate the contents of the returned document using test /ats/core/conformance-success.
    """
    assert (
        conformance_page.status_code == 200
    ), f"{conformance_page.url.path} did not return a status code of 200"


def test_ast8(
    schema_dir: Path, conformance_page: httpx.Response, conformance_classes: list
):
    """/ats/core/conformance-success
    Test Purpose
        Validate that the Conformance Declaration response complies with the required structure and contents.

    Test Method
        - Validate the response document against OpenAPI 3.0 schema confClasses.yaml
        - Validate that the document includes the conformance class "http://www.opengis.net/spec/ogcapi-features-1/1.0/conf/core"
        - Validate that the document list all OGC API conformance classes that the API implements.
    """
    with open(schema_dir / "confClasses.yaml") as file:
        schema = yaml.safe_load(file)
    response_json = conformance_page.json()
    requirement_class = "http://www.opengis.net/spec/ogcapi-features-1/1.0/conf/core"
    validate(response_json, schema)
    assert (
        requirement_class in conformance_classes
    ), f"OGC Features API Core requirement class ({requirement_class}) is not listed in the conformsTo property."



from pathlib import Path

import httpx
import yaml
from openapi_schema_validator import OAS30Validator, validate
from referencing import Registry

from .utils import get_links_by_rel, check_links_rel_type


def test_ast9(collections_page: httpx.Response):
    """/ats/core/fc-md-op
    Test Purpose
        Validate that information about the Collections can be retrieved from the expected location.

    Test Method
        - Issue an HTTP GET request to the URL {root}/collections
        - Validate that a document was returned with a status code 200
        - Validate the contents of the returned document using test /ats/core/fc-md-success.
    """
    assert (
        collections_page.status_code == 200
    ), f"{collections_page.url.path} did not return 200"


def test_ast10(
    collections_page: httpx.Response,
    collections: list[dict],
    schema_dir: Path,
    registry: Registry,
):
    """/ats/core/fc-md-success
    Test Purpose
        Validate that the Collections content complies with the required structure and contents.

    Test Method
        - Validate that all response documents comply with /ats/core/fc-md-links
        - Validate that all response documents comply with /ats/core/fc-md-items
        - In case the response includes a "crs" property, validate that the first value is either "http://www.opengis.net/def/crs/OGC/1.3/CRS84" or "http://www.opengis.net/def/crs/OGC/0/CRS84h"
        - Validate the collections content for all supported media types using the resources and tests identified in Schema and Tests for Collections content
    """
    assert (
        get_links_by_rel(collections_page, "self") != []
    ), f"{collections_page.url.path} must contain a link with rel == self"
    assert check_links_rel_type(
        collections_page
    ), f"all links given by {collections_page.url.path} must specify the rel and type parameter"

    with open(schema_dir / "collections.yaml") as file:
        schema = yaml.safe_load(file)
    validate(collections_page.json(), schema, cls=OAS30Validator, registry=registry)

    crs_requirement_classes = [
        "http://www.opengis.net/def/crs/OGC/1.3/CRS84",
        "http://www.opengis.net/def/crs/OGC/1.3/CRS84H",
    ]
    for collection in collections:
        if collection.get("crs", None) is not None:
            assert (
                collection.get("crs")[0] in crs_requirement_classes
            ), f"The first value of crs for collection {collection} must be one of {crs_requirement_classes}"



import httpx
import enum
import pytest


@pytest.fixture(scope="module")
def successful_filter_execution_status_codes():
    return [200, 204]


@pytest.fixture(scope="module")
def unsuccessful_filter_execution_status_codes():
    return [400]


@pytest.fixture(scope="module")
def filter_langs():
    return [
        {"name": "cql2-text", "default": True},
        {"name": "cql2-json", "default": False},
    ]


def test_ct6(http_client: httpx.Client, successful_filter_execution_status_codes):
    """/conf/filter/get-conformance
    Test Purpose
        Check that the API declares support for the conformance class
    Test Method
        Given:
            n/a
        When:
            the request for the Conformance Declaration is executed
                method: GET
                path: {apiURI}/conformance
                header: Accept: application/json
                authentication, if authentication credentials are provided
        Then:
            assert successful execution (status code is "200", Content-Type header is "application/json");
            assert that $.conformsTo is a string array that includes the value "http://www.opengis.net/spec/ogcapi-features-3/1.0/conf/filter".
    """
    response = http_client.get("/conformance", headers={"Accept": "application/json"})
    assert response.status_code in successful_filter_execution_status_codes
    assert response.headers.get("Content-Type") == "application/json"
    conforms_to = response.json().get("conformsTo")
    assert isinstance(conforms_to, list)
    assert (
        "http://www.opengis.net/spec/ogcapi-features-3/1.0/conf/filter" in conforms_to
    )


def test_ct7(
    http_client: httpx.Client,
    filterable_resources: list[dict],
    filter_langs,
    successful_filter_execution_status_codes,
    unsuccessful_filter_execution_status_codes,
):
    """/conf/filter/filter-param
    Test Purpose
        Check that the query parameter filter is supported
    Test Method
        Given:
            test "get-queryables" was successful
            the list of filterable resources
            the sample queryable of every filterable resource
        When:
            a request for each resource that supports filtering is executed without a filter parameter
                method: GET
                path: {apiURI}/{pathToResource}
                header: Accept: {responseMediaType}
                authentication, if authentication credentials are provided
        Then:
            assert successful execution (the status code is in the list of acceptable status codes for a successful execution, Content-Type header is {responseMediaType});
            store the result as the unfiltered result of the resource.
        When:
            a request for each resource that supports filtering is executed with a valid filter expression
                method: GET
                path: {apiURI}/{pathToResource}
                query parameters (before percent encoding): filter-lang={filter-lang}&filter={filter-valid} where {queryable} in {filter-valid} is replaced by the sample queryable of the filterable resource
                header: Accept: {responseMediaType}
                authentication, if authentication credentials are provided
        Then:
            assert successful execution (the status code is in the list of acceptable status codes for a successful execution, Content-Type header is {responseMediaType});
            assert that each returned resource matches the filter expression.
        When:
            a request for each resource that supports filtering is executed with an invalid filter expression
                method: GET
                path: {apiURI}/{pathToResource}
                query parameters (before percent encoding): filter-lang={filter-lang}&filter={filter-invalid} where {queryable} in {filter-invalid} is replaced by the sample queryable of the filterable resource
                header: Accept: {responseMediaType}
                authentication, if authentication credentials are provided
        Then:
            assert unsuccessful execution (the status code is in the list of acceptable status codes for an unsuccessful execution).
    """
    assert filterable_resources
    for resource in filterable_resources:
        headers = {"Accept": resource["media_type"]}
        unfiltered_response = http_client.get(resource["url"], headers=headers)
        assert (
            unfiltered_response.status_code in successful_filter_execution_status_codes
        )
        resource["unfiltered_response"] = unfiltered_response
        queryable = resource["queryable"]
        for lang in filter_langs:
            params = {"filter-lang": lang, "filter": f"{queryable}=valid"}
            response = http_client.get(
                resource["url"], headers=headers, params=params
            )
            assert response.status_code in successful_filter_execution_status_codes
            data = response.json()
            for feature in data.get("features", []):
                assert feature.get(queryable, "") == queryable
            params = {"filter-lang": lang, "filter": f"{queryable}=invalid"}
            response = http_client.get(
                resource["url"], headers=headers, params=params
            )
            assert response.status_code in unsuccessful_filter_execution_status_codes


def test_ct8(http_client: httpx.Client, filterable_resources: list[dict]):
    """/conf/filter/filter-lang-default
    Test Purpose
        Check that the query parameter filter-lang default value is supported
    Test Method
        Given:
            test "get-queryables" was successful
            the list of filterable resources
            the queryables of every filterable resource
            the filter language {filter-lang} is the default filter language
        When:
            a request for each resource that supports filtering is executed with a valid filter expression
                method: GET
                path: {apiURI}/{pathToResource}
                query parameters (before percent encoding): filter={filter-valid} where {queryable} in {filter-valid} is replaced by the sample queryable of the collection
                header: Accept: {responseMediaType}
                authentication, if authentication credentials are provided
        Then:
            assert successful execution (the status code is in the list of acceptable status codes for a successful execution, Content-Type header is {responseMediaType});
            assert that each returned resource matches the filter expression.
        When:
            a request for each resource that supports filtering is executed with an invalid filter expression
                method: GET
                path: {apiURI}/{pathToResource}
                query parameters (before percent encoding): filter={filter-invalid} where {queryable} in {filter-invalid} is replaced by the sample queryable of the collection
                header: Accept: {responseMediaType}
                authentication, if authentication credentials are provided
        Then:
            assert unsuccessful execution (the status code is in the list of acceptable status codes for an unsuccessful execution).
    """
    for resource in filterable_resources:
        url = f"/collections/{resource['id']}/items"
        headers = {"Accept": "application/geo+json"}
        queryable = resource["queryable"]
        params = {"filter": f"{queryable['key']}='{queryable['value']}'"}
        response = http_client.get(url, headers=headers, params=params)
        assert response.status_code in [200, 204]
        assert response.headers.get("Content-Type") == "application/geo+json"
        data = response.json()
        for feature in data.get("features", []):
            assert feature.get(queryable["key"], "") == queryable["value"]
        invalid_params = {"filter": "ooga=booga"}
        invalid_response = http_client.get(url, headers=headers, params=invalid_params)
        assert invalid_response.status_code == 400


def test_ct9(http_client: httpx.Client, filterable_resources: list[dict]):
    """/conf/filter/expression-construction
    Test Purpose
        Check that unknown queryables are rejected, if this is declared in the Queryables resource
    Test Method
        Given:
            test "get-queryables" was successful
            the list of filterable resources, reduced to those where additionalProperties is false`
            the sample queryable of every filterable resource in the list
        When:
            a request for each resource is executed with a filter expression with an unsupported queryable
                method: GET
                path: {apiURI}/{pathToResource}
                query parameters (before percent encoding): filter-lang={filter-lang}&filter={filter-valid} where {queryable} in {filter-valid} is replaced by "this_is_not_a_queryable"
                header: Accept: {responseMediaType}
                authentication, if authentication credentials are provided
        Then:
            assert unsuccessful execution (the status code is in the list of acceptable status codes for an unsuccessful execution).
    """
    resources_with_additional_props = [
        resource
        for resource in filterable_resources
        if resource["additional_properties"]
    ]
    for resource in resources_with_additional_props:
        url = f"/collections/{resource['id']}/items"
        headers = {"Accept": "application/geo+json"}
        params = {"filter-lang": "cql2-text", "filter": "this_is_not_a_queryable"}
        response = http_client.get(url, headers=headers, params=params)
        assert response.status_code == 400


def test_ct10(http_client: httpx.Client, filterable_resources: list[dict]):
    """/conf/filter/filter-crs-wgs84
    Test Purpose
        Check that spatial predicates assume WGS84 by default
    Test Method
        Given:
            test "get-queryables" was successful
            the list of filterable resources with a spatial queryable
            the spatial queryable of each filterable resource
            the WGS84 bbox of the resources in each filterable resource
        When:
            a request for each filterable resource is executed with a filter expression with a spatial predicate
                method: GET
                path: {apiURI}/{pathToResource}
                query parameters (before percent encoding): filter-lang={filter-lang}&filter={bbox-filter} where {spatialQueryable} in {bbox-filter} is replaced by by the spatial queryable, {x1} is replaced by the west-bound longitude of the WGS84 bbox of the resource, {y1} by the south-bound latitude, {x2} by the east-bound longitude, and {y2} by the north-bound latitude
                header: Accept: {responseMediaType}
                authentication, if authentication credentials are provided
        Then:
            assert successful execution (the status code is in the list of acceptable status codes for a successful execution, Content-Type header is {responseMediaType}).
            assert that result contains the same features as the unfiltered result of the filterable resource.
        When:
            a request for each filterable resource with a filter expression with a spatial predicate
                method: GET
                path: {apiURI}/{pathToResource}
                query parameters (before percent encoding): filter-lang={filter-lang}&filter={bbox-filter} where {spatialQueryable} in {bbox-filter} is replaced by by the spatial queryable, {x1} is replaced by "1000000", {y1} by "1000000", {x2} by "2000000", and {y2} by "2000000"
                header: Accept: {responseMediaType}
                authentication, if authentication credentials are provided
        Then:
            assert unsuccessful execution (the status code is in the list of acceptable status codes for an unsuccessful execution).
    """
    for resource in filterable_resources:
        url = f"/collections/{resource['id']}/items"
        headers = {"Accept": "application/geo+json"}
        response = http_client.get(
            f"/collections/{resource['id']}", headers={"Accept": "application/json"}
        )
        bbox = response.json()["extent"]["spatial"]["bbox"]
        cql_query = """
          "op": "s_within",
          "args": [
            { "property": {} },
            { "bbox": {} }
        """.format(
            resource["spatial_queryable"], bbox
        )
        params = {
            "filter-lang": CQL2_FILTER_LANG.json,
            "filter": f"{resource['spatial_queryable']}={bbox}",
        }



import warnings
from pathlib import Path

import yaml
from openapi_schema_validator import validate
from referencing import Registry

from .utils import get_links_by_rel, check_links_rel_type


def test_ast27(feature_responses: list[dict]):
    """/ats/core/f-op
    Test Purpose
        Validate that a feature can be retrieved from the expected location.

    Test Method
        - For a sufficiently large subset of all features in a feature collection (path /collections/{collectionId}),
          issue an HTTP GET request to the URL /collections/{collectionId}/items/{featureId} where {collectionId}
          is the id property for the collection and {featureId} is the id property of the feature.
        - Validate that a feature was returned with a status code 200
        - Validate the contents of the returned feature using test /ats/core/f-success
    """
    for response in feature_responses:
        assert (
            response["response"].status_code == 200
        ), f"feature {response['feature_id']}, did not return 200 at {response['response'].url.path} (status code: {response['response'].status_code})"


def test_ast28(feature_responses: list[dict], schema_dir: Path, registry: Registry):
    """/ats/core/f-success
    Test Purpose
        Validate that the Feature complies with the required structure and contents.

    Test Method
        - Validate that the Feature includes all required link properties using /ats/core/f-links
        - Validate the Feature for all supported media types using the resources and tests
          identified in Schema and Tests for Features
    """
    with open(schema_dir / "featureGeoJSON.yaml") as file:
        schema = yaml.safe_load(file)
    for response in feature_responses:
        feature = response["response"].json()
        if not feature:
            warnings.warn(
                f"No feature returned for feature {response['feature_id']} in collection {response['collection_id']}"
            )
            break
        validate(feature, schema, registry=registry)


def test_ast29(feature_responses):
    """/ats/core/f-links
    Test Purpose
        Validate that the required links are included in a Feature.

    Test Method
        Verify that the returned Feature includes:
            - a link to this response document (relation: self),
            - a link to the response document in every other media type supported by the server (relation: alternate).
            - a link to the feature collection that contains this feature (relation: collection).
        Verify that all links include the rel and type link parameters.
    """
    for response in feature_responses:
        assert (
            get_links_by_rel(response["response"], "self")
            and get_links_by_rel(response["response"], "collection")
            and check_links_rel_type(response["response"])
        ), f"Response for feature {response['feature_id']} in collection {response['collection_id']} must contain a link to itself and the parent collection and all links must specify the rel and type parameters."



import httpx


def test_ast1(http_client: httpx.Client):
    """/ats/core/http
    Test Purpose
        Validate that the resource paths advertised through the API conform with HTTP 1.1 and, where approprate, TLS.

    Test Method
        - All compliance tests shall be configured to use the HTTP 1.1 protocol exclusively.
        - For APIs which support HTTPS, all compliance tests shall be configured to use HTTP over
          TLS (RFC 2818) with their HTTP 1.1 protocol.
    """

    def check_http_version(response: httpx.Response):
        assert response.http_version == "HTTP/1.1"

    http_client.event_hooks["response"] = [check_http_version]



def test_ast11(collection_responses: list[dict]):
    """/ats/core/sfc-md-op
    Test Purpose
        Validate that the Collection content can be retrieved from the
        expected location.

    Test Method
        For every Feature Collection described in the Collections content,
        - issue an HTTP GET request to the URL /collections/{collectionId} where
        {collectionId} is the id property for the collection.
        - Validate that a Collection was returned with a status code 200
        - Validate the contents of the returned document using test /ats/core/sfc-md-success.
    """
    for response in collection_responses:
        status_code = response["response"].status_code
        assert (
            status_code == 200
        ), f"/collections/{response['id']} responded with {status_code}, expected 200"


def test_ast12(collections: list[dict], collection_responses: list[dict]):
    """/ats/core/sfc-md-success
    Test Purpose
        Validate that the Collection content complies with the required structure and contents.

    Test Method
        Verify that the content of the response is consistent with the content for this
        Feature Collection in the /collections response. That is, the values for
        id, title, description and extent are identical.
    """
    for collection in collections:
        collectionId = collection["id"]
        response = [
            response["response"].json()
            for response in collection_responses
            if response["id"] == collectionId
        ][0]
        assert all(
            (
                collection.get("id") == response.get("id"),
                collection.get("title") == response.get("title"),
                collection.get("description") == response.get("description"),
                collection.get("extent") == response.get("extent"),
            ),
        ), f"The properties of the collection '{collectionId}' do not match what is stated at /collections"



import os
import pytest
import httpx

try:
    from dotenv import load_dotenv
except ModuleNotFoundError:
    pass


@pytest.fixture(scope="session")
def instance_url() -> str:
    try:
        load_dotenv()
    except NameError:
        pass
    return os.environ.get("INSTANCE_URL", "")


@pytest.fixture(scope="session")
def http_client(instance_url: str) -> httpx.Client:
    with httpx.Client() as client:
        client.base_url = instance_url
        yield client


@pytest.fixture(scope="session", autouse=True)
def suite_props(record_testsuite_property, instance_url):
    record_testsuite_property("instance-url", instance_url)


if os.getenv("OGC_CALLED_FROM_PREZ"):
    pytest_plugins = "tests.test_ogc"



import re
from json import JSONDecodeError

import httpx

from .utils import parse_link_header


def test_ct1(http_client: httpx.Client):
    """/conf/queryables/get-conformance
    Test Purpose
        Check that the API declares support for the conformance class
    Test Method
        Given:
            n/a
        When:
            the request for the Conformance Declaration is executed
                method: GET
                path: {apiURI}/conformance
                authentication, if authentication credentials are provided
        Then:
            assert successful execution (status code is "200", Content-Type header is "application/json");
            assert that $.conformsTo is a string array that includes the value "http://www.opengis.net/spec/ogcapi-features-3/1.0/conf/queryables".
    """
    response = http_client.get("/conformance", headers={"Accept": "application/json"})
    assert response.status_code == 200, "Conformance endpoint did not return 200"
    assert response.headers.get("Content-Type") == "application/json", "Conformance endpoint did not return json"
    conforms_to = response.json().get("conformsTo")
    assert isinstance(conforms_to, list), "The conformsTo property is not an array"
    assert (
        "http://www.opengis.net/spec/ogcapi-features-3/1.0/conf/queryables"
        in conforms_to
    ), "http://www.opengis.net/spec/ogcapi-features-3/1.0/conf/queryables not listed in the conformsTo property."


def test_ct2(http_client: httpx.Client, filterable_resources: list[dict]):
    """/conf/queryables/get-queryables-uris
    Test Purpose
        Check that a link to the Queryables resource exists for every filterable resource
    Test Method
        Given:
            the list of filterable resources ({apiURI}/{pathToResource});
        When:
            a request is executed for every filterable resource
                method: HEAD (if HEAD results in a 405 response, use GET instead)
                path: {apiURI}/{pathToResource}
                header: Accept: application/json
                authentication, if authentication credentials are provided
        Then:
            assert successful execution (status code is "200");
            assert that the response includes a Link header with rel set to http://www.opengis.net/def/rel/ogc/1.0/queryables or [ogc-rel:queryables];
            store the href value as the Queryables URI for the filterable resource ({queryablesUri}).
    """
    assert filterable_resources
    for resource in filterable_resources:
        headers = {"Accept": resource["media_type"]}
        response = http_client.head(resource["url"], headers=headers)
        if response.status_code == 405:
            response = http_client.get(resource["url"], headers=headers)
        assert response.status_code == 200, f"resource {resource['id']} could not be retrieved."
        links = parse_link_header(response.headers.get("link", ""))
        rel = "http://www.opengis.net/def/rel/ogc/1.0/queryables"
        altrel = "[ogc-rel:queryables]"
        queryables_uri = [
            link.get("href")
            for link in links
            if link.get("rel") == rel or link.get("rel") == altrel
        ]
        assert queryables_uri, "No link with relation [ogc-rel:queryables] or http://www.opengis.net/def/rel/ogc/1.0/queryables available in the link header"
        resource["queryables_uri"] = queryables_uri


def test_ct3(http_client: httpx.Client, filterable_resources: list[dict]):
    """/conf/queryables/get-queryables
    Test Purpose
        Check that the Queryables resource exists for every filterable resource
    Test Method
        Given:
            test "get-queryables-uris" was successful
            the list of Queryables URI for the filterable resource ({queryablesUri})
        When:
            the request for the Queryables page is executed for every filterable resource
                method: GET
                path: {queryablesUri}
                header: Accept: application/schema+json
                authentication, if authentication credentials are provided
        Then:
            assert successful execution (status code is "200", Content-Type header is "application/schema+json");
            assert that the value of the $schema member is "https://json-schema.org/draft/2019-09/schema" or "https://json-schema.org/draft/2020-12/schema"
            assert that the value of the $id member is "{queryablesUri}".
            assert that the value of the type member is "object".
            assert that $.properties is a non-empty object;
            assert that each member in $.properties has an object as its value and the object either includes type member or a format member whose value starts with geometry-;
            assert that the response is a valid JSON Schema;
            store the key of an arbitrary property with a type member as the sample queryable of the filterable resource;
            store the key of an arbitrary property of the object as the spatial queryable of the filterable resource, if the value of member is an object that includes no type member and a format member with a value geometry-{type} where {type} is one of "point", "multipoint", "linestring", "multilinestring", "polygon", "multipolygon", or "geometry";
            store the value of the additionalProperties member or true, if it is not provided.
    """
    assert filterable_resources
    for resource in filterable_resources:
        queryables_uri = resource.get("queryables_uri")
        assert queryables_uri, f"No queryables_uri for collection: {resource['id']}"
        response = http_client.get(
            url=queryables_uri, headers={"Accept": "application/schema+json"}
        )
        assert response.status_code == 200, f"collection: {resource['id']} did not respond with 200"
        assert response.headers.get("Content-Type") == "application/schema+json", f"collection: {resource['id']} did not return application/schema+json"
        try:
            data = response.json()
        except JSONDecodeError:
            raise AssertionError("JSON Schema not valid")
        expected_schemas = [
            "https://json-schema.org/draft/2019-09/schema",
            "https://json-schema.org/draft/2020-12/schema",
        ]
        assert data.get("$schema") in expected_schemas, f"$schema is not one of the expected schemas: {resource['id']}"
        assert data.get("$id") == queryables_uri, f"$id not in response body at collection: {resource['id']}"
        assert data.get("type") == "object", f"type property is not equal to 'object', at collection: {resource['id']}"
        properties = data.get("properties", [])
        assert len(properties) > 0, f"no properties given for collection: {resource['id']}"
        queryable = None
        spatial_queryable = None
        for prop_id, prop in properties.items():
            assert isinstance(prop, dict)
            keys = prop.keys()
            assert "type" in keys or "format" in keys, f"missing type and format on property {prop_id} for collection: {resource['id']}"
            if re.match(
                "geometry-(point|multipoint|linestring|multilinestring|polygon|multipolygon|geometry)",
                prop.get("format", ""),
            ):
                spatial_queryable = prop_id
            else:
                queryable = prop_id
        additional_properties = data.get("additionalProperties", True)
        resource["queryable"] = queryable
        resource["spatial_queryable"] = spatial_queryable
        resource["additional_properties"] = additional_properties



import pytest
import httpx


@pytest.fixture(scope="module")
def successful_filter_execution_status_codes() -> list:
    return [200, 204]


@pytest.fixture(scope="module")
def unsuccessful_filter_execution_status_codes() -> list:
    return [400]


@pytest.fixture(scope="module")
def valid_queryable() -> str:
    return "valid"


@pytest.fixture(scope="module")
def invalid_queryable() -> str:
    return "invalid"


def test_ct4(http_client: httpx.Client, filterable_resources: list[dict]):
    """/conf/queryables-query-parameters/get-conformance
    Test Purpose
        Check that the API declares support for the conformance class
    Test Method
        Given:
            n/a
        When:
            the request for the Conformance Declaration is executed
                method: GET
                path: {apiURI}/conformance
                header: Accept: application/json
                authentication, if authentication credentials are provided
        Then:
            assert successful execution (status code is "200", Content-Type header is "application/json");
            assert that $.conformsTo is a string array that includes the value "http://www.opengis.net/spec/ogcapi-features-3/1.0/conf/queryables-query-parameters".
    """
    response = http_client.get("/conformance", headers={"Accept": "application/json"})
    assert response.status_code == 200, "/conformance did not return 200"
    assert response.headers.get("Content-Type") == "application/json"
    conforms_to = response.json().get("conformsTo")
    assert isinstance(conforms_to, list)
    assert (
        "http://www.opengis.net/spec/ogcapi-features-3/1.0/conf/queryables-query-parameters"
        in conforms_to
    )


def test_ct5(
    http_client: httpx.Client,
    filterable_resources: list[dict],
    valid_queryable: str,
    invalid_queryable: str,
    successful_filter_execution_status_codes: list,
    unsuccessful_filter_execution_status_codes: list,
):
    """/conf/queryables-query-parameters/query-param
    Test Purpose
        Check that query parameters for queryables is supported
    Test Method
        Given:
            test "get-queryables" was successful
            the list of collections
            the sample queryable of every collection
        When:
            a request for every filterable resource that supports filtering is executed and every queryable (queryable) with a valid value for the queryable ({valid-value})
                method: GET
                path: {apiURI}/collections/{collectionId}/items
                query parameters (before percent encoding): {queryable}={valid-value}
                header: Accept: {responseMediaType}
                authentication, if authentication credentials are provided
        Then:
            assert successful execution (the status code is in the list of acceptable status codes for a successful execution, Content-Type header is {responseMediaType});
            assert that each returned resource matches the filter.
        When:
            a request for every filterable resource that supports filtering is executed and every queryable (queryable) with a invalid value for the queryable ({invalid-value})
                method: GET
                path: {apiURI}/collections/{collectionId}/items
                query parameters (before percent encoding): {queryable}={invalid-value}
                header: Accept: {responseMediaType}
                authentication, if authentication credentials are provided
        Then:
            assert unsuccessful execution (the status code is in the list of acceptable status codes for an unsuccessful execution).
    """
    for resource in filterable_resources:
        media_type = resource["media_type"]
        headers = {"Accept": media_type}
        valid_params = {resource["queryable"]: "valid-value"}
        response = http_client.get(
            resource["url"], headers=headers, params=valid_params
        )
        assert response.status_code in successful_filter_execution_status_codes
        invalid_params = {resource["queryable"]: "[invalid-value]"}
        response = http_client.get(
            resource["url"], headers=headers, params=invalid_params
        )
        assert response.status_code in unsuccessful_filter_execution_status_codes



import os
from pathlib import Path

import pytest

root_dir = Path(__file__).parent


def run_features_core(instance_url: str, test_client):
    os.environ["INSTANCE_URL"] = instance_url
    os.environ["OGC_CALLED_FROM_PREZ"] = True
    pytest.main(args=[str(root_dir / "features/core")])



import json
import warnings
from datetime import datetime
from pathlib import Path

import geopandas as gpd
import httpx
import pytest
import yaml
from openapi_schema_validator import validate
from referencing import Registry
from shapely.geometry import Polygon

from .test_crs import assert_valid_feature_crs
from .utils import (
    get_links_by_rel,
    check_links_rel_type,
    get_api_parameter,
    get_features_responses,
)


@pytest.fixture(scope="module")
def bbox_responses() -> list:
    return []


def assert_bbox_in_range(feature: dict, bbox: str):
    ll_lon, ll_lat, ur_lon, ur_lat = bbox.split(",")
    corners = [(ll_lon, ll_lat), (ur_lon, ll_lat), (ur_lon, ur_lat), (ll_lon, ur_lat)]
    polygon = Polygon(corners)
    feature_gdf = gpd.read_file(json.dumps(feature), driver="GeoJSON")
    assert feature_gdf.intersects(polygon)[
        0
    ], f"""
        Returned features must intersect with the bounding box specified in the bbox parameter.
        bbox: {polygon}.
        feature: (id: {feature['id']}). (geometries: {feature_gdf.geometry})
        """.strip()


def test_ast13(features_responses: list[dict]):
    """/ats/core/fc-op
    Test Purpose
        Validate that features can be identified and extracted from a Collection using query parameters.

    Test Method
        For every feature collection identified in Collections, issue an HTTP GET request to the URL /collections/{collectionId}/items
        where {collectionId} is the id property for a Collection described in the Collections content.
        - Validate that a document was returned with a status code 200.
        - Validate the contents of the returned document using test /ats/core/fc-response.
        - Repeat these tests using the following parameter tests:
            - Bounding Box:
                Parameter /ats/core/fc-bbox-definition
                Response /ats/core/fc-bbox-response
            - Limit:
                Parameter /ats/core/fc-limit-definition
                Response /ats/core/fc-limit-response
            - DateTime:
                Parameter /ats/core/fc-time-definition
                Response /ats/core/fc-time-response
        Error conditions:
            Query Invalid /ats/core/query-param-invalid
            Query Unknown /ats/core/query-param-unknown
        Execute requests with combinations of the "bbox" and "datetime" query parameters and verify that only
        features are returned that match both selection criteria.
    """
    for response in features_responses:
        status_code = response["response"].status_code
        assert (
            status_code == 200
        ), f"/collections/{response['id']}/items responded with {status_code}, expected 200"


def test_ast14(
    collections: list[dict],
    api_model: dict,
):
    """/ats/core/fc-bbox-definition
    Test Purpose
        Validate that the bounding box query parameters are constructed correctly.

    Test Method
        - Verify that the bbox query parameter complies with the following definition
          (using an OpenAPI Specification 3.0 fragment):

                name: bbox
                in: query
                required: false
                schema:
                  type: array
                  minItems: 4
                  maxItems: 6
                  items:
                    type: number
                style: form
                explode: false

        - Use a bounding box with four numbers in all requests:
            Lower left corner, WGS 84 longitude
            Lower left corner, WGS 84 latitude
            Upper right corner, WGS 84 longitude
            Upper right corner, WGS 84 latitude
    """
    for collection in collections:
        path = f"/collections/{collection['id']}/items"
        bbox_def = get_api_parameter(api_model, path, "bbox")
        if not bbox_def:
            warnings.warn(f"bbox parameter missing from {path}")
        else:
            assert (
                bbox_def.get("in") == "query"
                and not bbox_def.get("required")
                and bbox_def.get("schema").get("type") == "array"
                and bbox_def.get("schema").get("minItems") == 4
                and bbox_def.get("schema").get("maxItems") == 6
                and bbox_def.get("schema").get("items").get("type") == "number"
                and bbox_def.get("style") == "form"
                and not bbox_def.get("explode")
            ), f"bbox parameter ill defined at {path}"


def test_ast15(
    collections: list[dict],
    http_client: httpx.Client,
    bbox_responses: list,
    bbox: str = "5,48,20,70",
):
    """/ats/core/fc-bbox-response
    Test purpose
        Validate that the bounding box query parameters are processed corrrectly.

    Test Method
        - Verify that only features that have a spatial geometry that intersects the bounding box
          are returned as part of the result set.
        - Verify that the bbox parameter matched all features in the collection that were not
          associated with a spatial geometry (this is only applicable for datasets that include
          features without a spatial geometry).
        - Verify that the coordinate reference system of the geometries is WGS 84 longitude/latitude
          ("http://www.opengis.net/def/crs/OGC/1.3/CRS84" or "http://www.opengis.net/def/crs/OGC/0/CRS84h")
          since no parameter bbox-crs was specified in the request.
    """
    if not bbox_responses:
        bbox_responses = get_features_responses(
            collections, http_client, params={"bbox": bbox}
        )
    for response in bbox_responses:
        data = response["response"].json()
        for i, feature in enumerate(data.get("features", [])):
            # as per official test suite. stop after checking 100 features
            if i > 100:
                break
            assert_bbox_in_range(feature, bbox)
            assert_valid_feature_crs(feature)


def test_ast16(collections: list[dict], api_model: dict):
    """/ats/core/fc-limit-definition
    Test Purpose
        Validate that the limit query parameters are constructed corrrectly.

    Test Method
        - Verify that the limit query parameter complies with the following definition
          (using an OpenAPI Specification 3.0 fragment):

                name: limit
                in: query
                required: false
                schema:
                  type: integer
                style: form
                explode: false
        - Note that the API can define values for "minimum", "maximum" and "default".
    """
    for collection in collections:
        path = f"/collections/{collection['id']}/items"
        limit_def = get_api_parameter(api_model, path, "limit")
        if not limit_def:
            warnings.warn(f"limit parameter missing from {path}")
        else:
            assert (
                limit_def.get("in") == "query"
                and not limit_def.get("required")
                and limit_def.get("schema").get("type") == "integer"
                and limit_def.get("style") == "form"
                and not limit_def.get("explode")
            ), f"limit parameter ill defined at {path}"


def test_ast17(collections: list[dict], http_client: httpx.Client, limit: int = 5):
    """/ats/core/fc-limit-response
    Test Purpose
        Validate that the limit query parameters are processed correctly.

    Test Method
        - Count the Features which are on the first level of the collection. Any nested objects contained
          within the explicitly requested items are not be counted.
        - Verify that this count is not greater than the value specified by the limit parameter.
        - If the API definition specifies a maximum value for limit parameter, verify that the count does
          not exceed this maximum value.
    """
    features_responses = get_features_responses(
        collections, http_client, params={"limit": limit}
    )
    for response in features_responses:
        features = response["response"].json().get("features", [])
        if not features:
            warnings.warn(f"No features returned from {response['response'].url.path}")
            break
        assert (
            len(features) <= limit
        ), f"Number of features exceeds the limit parameter value, limit: {limit}, received {len(features)}"


def test_ast18(collections: list[dict], api_model: dict):
    """/ats/core/fc-time-definition
    Test Purpose
        Validate that the dateTime query parameters are constructed correctly.

    Test Method
        - Verify that the datetime query parameter complies with the following definition
          (using an OpenAPI Specification 3.0 fragment):

            name: datetime
            in: query
            required: false
            schema:
                type: string
            style: form
            explode: false
    """
    for collection in collections:
        path = f"/collections/{collection['id']}/items"
        datetime_def = get_api_parameter(api_model, path, "datetime")
        if not datetime_def:
            warnings.warn(f"datetime parameter missing from {path}")
        else:
            assert (
                datetime_def.get("in") == "query"
                and not datetime_def.get("required")
                and datetime_def.get("schema").get("type") == "string"
                and datetime_def.get("style") == "form"
                and not datetime_def.get("explode")
            ), f"datetime parameter ill defined at {path}"


def test_ast19(collections: list[dict], http_client: httpx.Client):
    """/ats/core/fc-time-response
    Test Purpose
        Validate that the dataTime query parameters are processed correctly.

    Test Method
        - Verify that only features that have a temporal geometry that intersects the temporal information
          in the datetime parameter were included in the result set
        - Verify that all features in the collection that are not associated with a temporal geometry are
          included in the result set
        - Validate that the datetime parameter complies with the syntax described in /req/core/fc-time-response.
    """
    date_time = "2015-01-01/2024-01-01"
    responses = get_features_responses(
        collections, http_client, params={"dateTime": date_time}
    )
    for response in responses:
        # This test not implemented in the official test suite either
        assert response


def test_ast22(
    schema_dir: Path,
    registry: Registry,
    features_responses: list[dict],
):
    """/ats/core/fc-response
    Test Purpose
        Validate that the Feature Collections complies with the require structure and contents.

    Test Method
        - Validate that the type property is present and has a value of FeatureCollection
        - Validate the features property is present and that it is populated with an array of feature items.
        - Validate that only Features which match the selection criteria are included in the Feature Collection.
        - If the links property is present, validate that all entries comply with /ats/core/fc-links
        - If the timeStamp property is present, validate that it complies with /ats/core/fc-timeStamp
        - If the numberMatched property is present, validate that it complies with /ats/core/fc-numberMatched
        - If the numberReturned property is present, validate that it complies with /ats/core/fc-numberReturned
        - Validate the collections content for all supported media types using the resources and tests identified in
          Schema and Tests for Feature Collections
    """
    with open(schema_dir / "featureCollectionGeoJSON.yaml") as file:
        schema = yaml.safe_load(file)
    for response in features_responses:
        assert (
            response["response"].json().get("features", [])
        ), f"No features listed at /collections/{response['id']}/items"
        validate(response["response"].json(), schema, registry=registry)


def test_ast23(features_responses: list[dict]):
    """/ats/core/fc-links
    Test Purpose
        Validate that the required links are included in the Collections document.

    Test Method
        Verify that the response document includes:
            - a link to this response document (relation: self),
            - a link to the response document in every other media type supported by the server (relation: alternate).
            - Verify that all links include the rel and type link parameters.
    """
    for response in features_responses:
        assert get_links_by_rel(
            response["response"], "self"
        ), f"The response from /collections/{response['id']}/items does not contain a link to itself."
        assert check_links_rel_type(
            response["response"]
        ), f"Not all links contain a rel and type parameter at /collections/{response['id']}/items"


def test_ast24(
    features_responses: list[dict],
):
    """/ats/core/fc-timeStamp
    Test Purpose
        Validate the timeStamp parameter returned with a Features response

    Test Method
        Validate that the timeStamp value is set to the time when the response was generated.
    """
    for response in features_responses:
        ts_before = response["before"]
        ts_after = response["after"]
        try:
            ts = datetime.fromisoformat(
                response["response"].json().get("timeStamp", "")
            )
        except ValueError:
            warnings.warn(f"No timestamp given at {response['id']}")
            break
        assert (
            ts_before <= ts <= ts_after
        ), f"unexpected value for timeStamp. Expected: {ts_before:%Y-%m-%dT%H:%M:%S} <= {ts:%Y-%m-%dT%H:%M:%S} <= {ts_after:%Y-%m-%dT%H:%M:%S}"


def test_ast25(
    collections: list[dict],
    http_client: httpx.Client,
    bbox_responses: list,
    bbox: str = "5,48,20,70",
):
    """/ats/core/fc-numberMatched
    Test Purpose
        Validate the numberMatched parameter returned with a Features response

    Test Method
        Validate that the value of the numberMatched parameter is identical to the number of features
        in the feature collections that match the selection parameters like bbox, datetime or
        additional filter parameters.
    """
    if not bbox_responses:
        bbox_responses = get_features_responses(
            collections, http_client, params={"bbox": bbox}
        )
    for response in bbox_responses:
        data = response["response"].json()
        numberMatched = data.get("numberMatched", -1)
        if numberMatched == -1:
            warnings.warn(f"numberMatched parameter not given at {response['id']}")
            break
        features = data.get("features", [])
        nxt_link = get_links_by_rel(response["response"], "next", only_first=True)
        total_returned = len(features)
        while nxt_link:
            nxt_page = http_client.get(
                url=nxt_link["href"], headers={"Accept": "application/geo+json"}
            )
            features = nxt_page.json().get("features", [])
            nxt_link = get_links_by_rel(nxt_page, "next", only_first=True)
            total_returned += len(features)
        assert (
            total_returned == numberMatched
        ), f"numberMatched parameter claims {numberMatched} features. but {total_returned} were found."


def test_ast26(
    collections: list[dict],
    http_client: httpx.Client,
    bbox_responses: list,
    bbox: str = "5,48,20,70",
):
    """/ats/core/fc-numberReturned
    Test Purpose
        Validate the numberReturned parameter returned with a Features response

    Test Method
        Validate that the numberReturned value is identical to the number of features in the response
    """

    if not bbox_responses:
        bbox_responses = get_features_responses(
            collections, http_client, params={"bbox": bbox}
        )
    for response in bbox_responses:
        data = response["response"].json()
        numberReturned = data.get("numberReturned", -1)
        if numberReturned == -1:
            warnings.warn(f"numberReturned parameter not given at {response['id']}")
            break
        total_returned = len(data.get("features", []))
        assert (
            total_returned == numberReturned
        ), f"numberReturned parameter claims {numberReturned} features. but got {total_returned}"



from json import JSONDecodeError
import pytest
import httpx

from .utils import get_supported_media_types


def pytest_collection_modifyitems(items):
    # Run the collected test functions in order of conformance test number
    items[:] = sorted(items, key=lambda item: int(item.name.replace("test_ct", "")))


@pytest.fixture(scope="session")
def implemented_media_types() -> set:
    return {"application/geo+json", ...}


@pytest.fixture(scope="session")
def collections(http_client) -> list[dict]:
    response = http_client.get("/collections", headers={"Accept": "application/json"})
    try:
        collections = response.json().get("collections", [])
    except JSONDecodeError:
        collections = []
    return collections


@pytest.fixture(scope="session")
def filterable_resources(
    collections: list[dict], http_client: httpx.Client, implemented_media_types: set
) -> list[dict]:
    resources = []
    for collection in collections:
        # Check for supported media types
        response = http_client.get(f"/collections/{collection['id']}/items")
        supported_media_types = get_supported_media_types(response)
        media_types = supported_media_types.intersection(implemented_media_types)
        if media_types:
            media_type = media_types.pop()  # Only keep one of the supported media types for testing
        else:
            raise Exception(
                f"None of the media types supported at {collection['id']} have been implemented in this test suite."
            )
        resource = dict()
        resource["id"] = collection["id"]
        resource["url"] = f"/collections/{collection['id']}/items"
        resource["media_type"] = media_type
        resources.append(resource)
    return resources
